#!/bin/python3
import sys
from functools import partial

import yaml
import torch
from tqdm import tqdm

from colourize.cli import parse_args
from colourize.colourspace import Colourspace, COLOURSPACES
from colourize import normalize
from colourize import serialize
from colourize import _select


def main():
    args = parse_args(sys.argv[1:])

    if args.all_colours:
        colourspaces = COLOURSPACES
    else:
        colourspaces = [Colourspace(colour) for colour in args.colours]

    if args.action == 'normalize':
        normalize.outer_loop(args.dataset_name,
                             args.root,
                             colourspaces,
                             args.outfile,
                             args.force)
    if args.action == 'serialize':
        serialize.outer_loop(args.dataset_name,
                             args.root,
                             colourspaces,
                             args.outdir,
                             args.force)
    if args.action == 'colourize':
        colourize(args.dataset_name,
                  args.root,
                  colourspaces,
                  args.outdir,
                  args.force)


def colourize(dataset_name, dataset_root, colourspaces, outdir, force):
    norms = {}
    outpath = outdir / (dataset_name + '.yml')
    if outdir and outpath.exists():
        print("Outfile exists, loading...")
        with open(outpath, 'r') as infile:
            norms = yaml.load(infile)

    for colourspace in colourspaces:
        colourdir = outdir / (dataset_name + '-' + colourspace.std_name)

        # FIXME(S): we prevent both not just one
        if colourdir.exists():
            print("Colourspace {} already serialized".format(colourspace))
            continue
        if colourspace.std_name in norms.keys() and not force:
            print("Normas for Colourspace {} already calculated".format(colourspace))
            continue

        traindir = colourdir / 'train'
        valdir = colourdir / 'val'
        traindir.mkdir(parents=True, exist_ok=True)
        valdir.mkdir(parents=True, exist_ok=True)
        print("Colourizing {} colourspace".format(colourspace))
        transforms = _select.transforms(colourspace.colour)
        loaders = _select.dataset(dataset_name, dataset_root, transforms)
        mean, std = process_colourspace(loaders,
                                        colourdir,
                                        colourspace.channels)
        mean, std = mean.tolist(), std.tolist()
        norms[colourspace.std_name] = {'mean': mean, 'std': std}

        print("{}:".format(colourspace))
        print("\tMean: {}\n\tStandard Deviation: {}".format(mean, std))
        if outpath:
            with open(outpath, 'w') as outfile:
                yaml.dump(norms, outfile, default_flow_style=False)

    if outpath:
        with open(outpath, 'w') as outfile:
            yaml.dump(norms, outfile, default_flow_style=False)


def process_colourspace(loaders, path, channels):
    """
    """
    cnt = 0
    fst_moment = torch.zeros(channels)
    snd_moment = torch.zeros(channels)

    train_loader, val_loader = loaders
    serialize_train = partial(serialize.serialize, path / 'train')
    serialize_val = partial(serialize.serialize, path / 'val')
    print("Training set:")
    for idx, data in enumerate(tqdm(train_loader)):
        image, _ = data
        serialize_train(idx, data)
        cnt, fst_moment, snd_moment = normalize.norm(image,
                                                     cnt,
                                                     fst_moment,
                                                     snd_moment)
    print("Validation set:")
    for idx, data in enumerate(tqdm(val_loader)):
        serialize_val(idx, data)

    return fst_moment, torch.sqrt(snd_moment - fst_moment ** 2)


if __name__ == '__main__':
    main()



"""Colourspace with metadata required to perform conversion.

Name for the ``colour-science`` conversion graph and channels.
"""
import warnings


class Colourspace():
    _two_ch = ['CIE UCS uv', 'CIE Luv uv']

    _three_ch = ['sRGB', 'CAM02SCD', 'CAM02LCD', 'CAM02UCS', 'CIECAM02 JMh',
                 'Scene-Referred RGB', 'IPT', 'hdr IPT', 'CMY', 'CIE LAB',
                 'hdr CIELab', 'CIE XYZ', 'Hunter Rdab', 'YCbCr', 'YcCbcCrc',
                 'ICtCp', 'CIE UCS', 'CIE LCHab', 'JzAzBz', 'HED', 'HSV',
                 'CIE xyY', 'OSA UCS', 'HSL', 'CIE UVW', 'Hunter LAB',
                 'CIE Luv', 'Munsell Colour']

    _four_ch = ['CMYK', 'Prismatic']

    _warn_non_fatal = ['HSV', 'CIE xyY', 'OSA UCS', 'HSL', 'CMYK', 'Prismatic']

    _fatal_nan = ['CIE UCS uv', 'CIE Luv uv', 'CIE UVW', 'Hunter LAB',
                  'CIE Luv']

    _fatal_scaling = ['Munsell Colour']

    def __init__(self, name, channels=None):
        self.colour = name
        self.std_name = name.replace(' ', '-').lower()
        if not channels:
            self.channels = self._ch_from_name()
        else:
            self.channels = channels

        if self.colour in self._warn_non_fatal:
            warnings.warn("Colourspace {} raises conversion errors \
                          from sRGB".format(self.colour))
        if self.colour in self._fatal_nan:
            warnings.warn("Colourspace {} conversion from sRGB \
                          produces NANs".format(self.colour))
        if self.colour in self._fatal_scaling:
            warnings.warn("Colourspace {} conversion from sRGB \
                          fails scaling".format(self.colour))

    def __str__(self):
        return self.colour

    def __eq__(self, other):
        return self.std_name == other.std_name

    def _ch_from_name(self):
        if self.colour in self._two_ch:
            return 2
        elif self.colour in self._three_ch:
            return 3
        elif self.colour in self._four_ch:
            return 4
        else:
            raise ValueError("Colourspace {} is not implemented.", self.colour)


with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    COLOURSPACES = [Colourspace('sRGB'),
                    Colourspace('CAM02SCD'),
                    Colourspace('CAM02LCD'),
                    Colourspace('CAM02UCS'),
                    Colourspace('CIECAM02 JMh'),
                    Colourspace('Scene-Referred RGB'),
                    Colourspace('IPT'),
                    Colourspace('hdr IPT'),
                    Colourspace('CMY'),
                    Colourspace('CIE LAB'),
                    Colourspace('hdr CIELab'),
                    Colourspace('CIE XYZ'),
                    Colourspace('Hunter Rdab'),
                    Colourspace('YCbCr'),
                    Colourspace('YcCbcCrc'),
                    Colourspace('ICtCp'),
                    Colourspace('CIE UCS'),
                    Colourspace('CIE LCHab'),
                    Colourspace('JzAzBz'),
                    Colourspace('HED'),              # Scikit-Image
                    Colourspace('HSV'),              # Non-Fatal Errors
                    Colourspace('CIE xyY'),          # Non-Fatal Errors
                    Colourspace('OSA UCS'),          # Non-Fatal Errors
                    Colourspace('HSL'),              # Non-Fatal Errors
                    Colourspace('CMYK'),             # Non-Fatal Errors
                    Colourspace('Prismatic')]        # Non-Fatal Errors



from torch.utils.data import DataLoader
from torchvision.datasets import CIFAR10
from torchvision.transforms import ToTensor, Compose

from colourize.transforms import ConvertColorspace, ToArray


def transforms(colorspace):
    """Return transfroms for preprocessing a colorspace.
    """
    transforms = []
    if colorspace == 'sRGB':
        transforms.append(ToTensor()) # FIXME: This only scales because of
                                      # ToTensor, which is now done in
                                      # colorspace conversion.
    else:
        transforms.append(ToArray())
        transforms.append(ConvertColorspace(colorspace))
        transforms.append(ToTensor())
    return Compose(transforms)


def dataset(name, root, transforms):
    """Return dataloader for dataset.
    """
    if name.lower() == 'cifar10':
        train_dataset = CIFAR10(root, transform=transforms, train=True)
        val_dataset = CIFAR10(root, transform=transforms, train=False)
    else:
        raise ValueError("Dataset {} not found")

    train_loader = DataLoader(
        train_dataset,
        batch_size=1,
        num_workers=10,
        shuffle=False,
    )

    val_loader = DataLoader(
        val_dataset,
        batch_size=1,
        num_workers=10,
        shuffle=False,
    )

    return train_loader, val_loader



"""
"""
import yaml
import torch
from tqdm import tqdm

from colourize import _select


def outer_loop(dataset_name, dataset_root, colourspaces, outfile, force):
    """
    """
    norms = {}
    if outfile and outfile.exists():
        print("Outfile exists, loading...")
        with open(outfile, 'r') as infile:
            norms = yaml.load(infile)

    for colourspace in colourspaces:
        if colourspace.std_name in norms.keys() and not force:
            print("Normas for Colourspace {} already calculated".format(colourspace))
            continue

        print("Calculating norms for {} colourspace".format(colourspace))
        transforms = _select.transforms(colourspace.colour)
        train_loader, _ = _select.dataset(dataset_name, dataset_root, transforms)
        mean, std = inner_loop(train_loader, colourspace.channels)
        mean, std = mean.tolist(), std.tolist()
        norms[colourspace.std_name] = {'mean': mean, 'std': std}

        print("{}:".format(colourspace))
        print("\tMean: {}\n\tStandard Deviation: {}".format(mean, std))

    if outfile:
        with open(outfile, 'w') as outfile:
            yaml.dump(norms, outfile, default_flow_style=False)


def inner_loop(loader, channels):
    """Inner loop over each colourspace.
    """
    cnt = 0
    fst_moment = torch.zeros(channels)
    snd_moment = torch.zeros(channels)

    for data, _ in tqdm(loader):
        cnt, fst_moment, snd_moment = norm(data, cnt, fst_moment, snd_moment)

    return fst_moment, torch.sqrt(snd_moment - fst_moment ** 2)


def norm(data, cnt, fst_moment, snd_moment):
    """Compute the mean and sd in an online fashion

        Var[x] = E[X^2] - E^2[X]
    """
    b, c, h, w = data.shape
    nb_pixels = b * h * w
    sum_ = torch.sum(data, dim=[0, 2, 3])
    sum_of_square = torch.sum(data ** 2, dim=[0, 2, 3])
    fst_moment = (cnt * fst_moment + sum_) / (cnt + nb_pixels)
    snd_moment = (cnt * snd_moment + sum_of_square) / (cnt + nb_pixels)
    cnt += nb_pixels
    return cnt, fst_moment, snd_moment



from pathlib import Path
from glob import glob

import torch
from torch.utils.data import Dataset


class ColourizedDataset(Dataset):
    """A PyTorch Dataset for preprocessed datasets.

    Colourized dataset, serialized to (image, target) paris.
    """
    def __init__(self, root, transforms=None, train=True):
        if train:
            self.root = Path(root).expanduser() / 'train'
        else:
            self.root = Path(root).expanduser() / 'val'

        if not self.root.exists():
            raise(ValueError("Path {} does not exist".format(self.root)))

        regex = str(self.root) + '/*'
        self.transforms = transforms
        self.data = glob(regex)

    def __getitem__(self, idx):
        image, target = torch.load(self.data[idx])
        image.squeeze_(0)
        if self.transforms:
            image = self.transforms(image)
        return image, target.item()

    def __len__(self):
        return len(self.data)



"""CLI interface for colorize.

Parses CLI argments and ensures they are valid.
"""
import argparse
from pathlib import Path


def parse_args(args):
    """Parse CLI arguments.
    """
    parent_parser = argparse.ArgumentParser(add_help=False)
    parent_parser.add_argument('dataset_name',
                        help='Dataset name')
    parent_parser.add_argument('root',
                        help='Dataset root location')
    parent_parser.add_argument('-f',
                        '--force',
                        action='store_true',
                        help='Overwrite norms/dataset if exists')
    colours = parent_parser.add_mutually_exclusive_group(required=True)
    colours.add_argument('--all',
                         dest='all_colours',
                         action='store_true',
                         help='Preprocess all colourspaces')
    colours.add_argument('--colours',
                         type=str,
                         nargs='+',
                         help='Colourspaces to process')

    description = "Colourize\n\n" \
                  "A tool for preprocessing image datasets."
    parser = argparse.ArgumentParser(description=description,
                                            add_help=False)


    subparsers = parser.add_subparsers(title='Subcommands',
                                       dest='action',
                                       required=True)


    parser_colourize_help = "Serialize and calculate normalization values " \
                            "for a dataset to disk as a PyTorch tensors and " \
                            "YAML for the specified colourspace."
    parser_colourize = subparsers.add_parser('colourize',
                                             parents=[parent_parser],
                                             help=parser_colourize_help)
    parser_colourize.add_argument('-o',
                                  '--outdir',
                                  default=Path.cwd() / 'colourized_data',
                                  type=Path,
                                  help='Path to directory where output should '
                                       'be written')

    parser_serialize_help = "Serialize a dataset to disk as a PyTorch " \
                            "tensor in the specified colourspace."
    parser_serialize = subparsers.add_parser('serialize',
                                             parents=[parent_parser],
                                             help=parser_serialize_help)
    parser_serialize.add_argument('-o',
                                  '--outdir',
                                  default=Path.cwd() / 'colourized_data',
                                  type=Path,
                                  help='Path to directory where output should '
                                       'be written')

    parser_normalize_help = "Calculate the normalization values for the " \
                            "given dataset and colourspace."
    parser_normalize = subparsers.add_parser('normalize',
                                             parents=[parent_parser],
                                             help=parser_normalize_help)
    parser_normalize.add_argument('-o',
                                  '--outfile',
                                  default=Path.cwd() / 'norms.yml',
                                  type=Path,
                                  help='Path to write normalization values.')
    parser_normalize.add_argument('--shuffle',
                                  action='store_true',
                                  help='Shuffle dataset')
    parser_normalize.add_argument('--seed',
                                  type=int,
                                  help='Random seed')
    parser_normalize.add_argument('--max',
                                  type=int,
                                  help='Max number of images to process')

    args = parser.parse_args(args)
    return args



from functools import partial

import torch
from tqdm import tqdm

from colourize import _select


def outer_loop(dataset_name, dataset_root, colourspaces, outdir, force):
    """
    """
    for colourspace in colourspaces:
        colourdir = outdir / (dataset_name + '-' + colourspace.std_name)
        if colourdir.exists():
            print("Colourspace {} already serialized".format(colourspace))
            continue
        colourdir.mkdir(parents=True, exist_ok=True)

        print("Serializing {} colourspace".format(colourspace))
        transforms = _select.transforms(colourspace.colour)
        train_loader, val_loader = _select.dataset(dataset_name,
                                                   dataset_root,
                                                   transforms)
        inner_loop(train_loader, colourdir / 'train')
        inner_loop(val_loader, colourdir / 'val')


def inner_loop(loader, path):
    serialize_ = partial(serialize, path)
    for idx, data in enumerate(tqdm(loader)):
        serialize_(idx, data)


def serialize(path, idx, data):
    path = path / '{0:010d}.pth'.format(idx)
    torch.save(data, path)



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['colourize']

package_data = \
{'': ['*']}

install_requires = \
['colour-science==0.3.14',
 'pyyaml>=5.1,<6.0',
 'scikit-image>=0.16.2,<0.17.0',
 'torch>=1.2.0,<2.0.0',
 'torchvision>=0.4.2,<0.5.0',
 'tqdm>=4.39,<5.0']

entry_points = \
{'console_scripts': ['clz = colourize.colourize:main']}

setup_kwargs = {
    'name': 'colourize',
    'version': '0.1.0',
    'description': 'Convert image datasets to other colourspaces.',
    'long_description': '## Colourize\n\n> Convert a dataset of images into many different colourspaces.\n\nA command line tool to preprocess a dataset into many different colourspaces.\nAlso provides a PyTorch `Dataset` that will work on the output and a PyTorch\ntransform to do the conversion online during training instead.\n\n### Currently in alpha this is just a MWE\n',
    'author': 'Derek Goddeau',
    'author_email': 'derek.j.goddeau@pm.me',
    'url': 'https://gitlab.com/starshell/datadyne/colourize',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import colour
import numpy as np

from skimage.color import rgb2hed
from colour.utilities import from_range_int, suppress_warnings
import torchvision.transforms.functional as F


class ConvertColorspace():
    """Convert RGB image to a given colorspace.

    Scales image to value range [0, 1].
    """
    def __init__(self, tospace, fromspace='sRGB'):
        self.fromspace = fromspace.lower()
        self.tospace = tospace.lower()

    def __call__(self, img):
        """
        Args:
            img (NumPy array):

        Returns:
        """
        if self.tospace == 'hed':
            img = rgb2hed(img)
        else:
            with colour.domain_range_scale('1'):
                img = from_range_int(img)
                with suppress_warnings(colour_usage_warnings=True):
                    img = colour.convert(img, self.fromspace, self.tospace)
        return np.float32(img)

    def __repr__(self):
        return self.__class__.__name__  # FIXME + '(p={})'.format(self.p)


class ToArray():
    """Convert a PIL Image to a NumPy array.
    """
    def __call__(self, img):
        return np.asarray(img)


class ToTensor():
    """Convert a ``PIL Image`` or ``numpy.ndarray`` to tensor.

    Converts a PIL Image or numpy.ndarray (H x W x C) in the range
    [0, 255] to a torch.FloatTensor of shape (C x H x W) in the range
    [0.0, 1.0] if the PIL Image belongs to one of the modes
    (L, LA, P, I, F, RGB, YCbCr, RGBA, CMYK, 1) or if the numpy.ndarray
    has dtype = np.uint8

    In the other cases, tensors are returned without scaling.
    """
    def __call__(self, img):
        """
        Args:
            img(PIL Image or numpy.ndarray): Image to be converted to tensor.

        Returns:
            Tensor: Converted image.
        """
        return F.to_tensor(img)

    def __repr__(self):
        return self.__class__.__name__ + '()'
